@@ -1,14 +1,15 @@
 # Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
-# For details: https://github.com/PyCQA/pylint/blob/master/LICENSE
+# For details: https://github.com/PyCQA/pylint/blob/main/LICENSE
 
 import collections
 import copy
 import itertools
 import tokenize
 from functools import reduce
-from typing import List
+from typing import Dict, Iterator, List, NamedTuple, Optional, Tuple, Union, cast
 
 import astroid
+from astroid.util import Uninferable
 
 from pylint import checkers, interfaces
 from pylint import utils as lint_utils
@@ -41,8 +42,6 @@ CALLS_RETURNING_CONTEXT_MANAGERS = frozenset(
         "tarfile.TarFile",
         "tarfile.TarFile.open",
         "multiprocessing.context.BaseContext.Pool",
-        "concurrent.futures.thread.ThreadPoolExecutor",
-        "concurrent.futures.process.ProcessPoolExecutor",
         "subprocess.Popen",
     )
 )
@@ -103,6 +102,77 @@ def _is_trailing_comma(tokens: List[tokenize.TokenInfo], index: int) -> bool:
     return False
 
 
+def _is_inside_context_manager(node: astroid.Call) -> bool:
+    frame = node.frame()
+    if not isinstance(
+        frame, (astroid.FunctionDef, astroid.BoundMethod, astroid.UnboundMethod)
+    ):
+        return False
+    return frame.name == "__enter__" or utils.decorated_with(
+        frame, "contextlib.contextmanager"
+    )
+
+
+def _is_part_of_with_items(node: astroid.Call) -> bool:
+    """
+    Checks if one of the node's parents is an ``astroid.With`` node and that the node itself is located
+    somewhere under its ``items``.
+    """
+    frame = node.frame()
+    current = node
+    while current != frame:
+        if isinstance(current, astroid.With):
+            items_start = current.items[0][0].lineno
+            items_end = current.items[-1][0].tolineno
+            return items_start <= node.lineno <= items_end
+        current = current.parent
+    return False
+
+
+def _will_be_released_automatically(node: astroid.Call) -> bool:
+    """Checks if a call that could be used in a ``with`` statement is used in an alternative
+    construct which would ensure that its __exit__ method is called."""
+    callables_taking_care_of_exit = frozenset(
+        (
+            "contextlib._BaseExitStack.enter_context",
+            "contextlib.ExitStack.enter_context",  # necessary for Python 3.6 compatibility
+        )
+    )
+    if not isinstance(node.parent, astroid.Call):
+        return False
+    func = utils.safe_infer(node.parent.func)
+    if not func:
+        return False
+    return func.qname() in callables_taking_care_of_exit
+
+
+class ConsiderUsingWithStack(NamedTuple):
+    """Stack for objects that may potentially trigger a R1732 message
+    if they are not used in a ``with`` block later on."""
+
+    module_scope: Dict[str, astroid.NodeNG] = {}
+    class_scope: Dict[str, astroid.NodeNG] = {}
+    function_scope: Dict[str, astroid.NodeNG] = {}
+
+    def __iter__(self) -> Iterator[Dict[str, astroid.NodeNG]]:
+        yield from (self.function_scope, self.class_scope, self.module_scope)
+
+    def get_stack_for_frame(
+        self, frame: Union[astroid.FunctionDef, astroid.ClassDef, astroid.Module]
+    ):
+        """Get the stack corresponding to the scope of the given frame."""
+        if isinstance(frame, astroid.FunctionDef):
+            return self.function_scope
+        if isinstance(frame, astroid.ClassDef):
+            return self.class_scope
+        return self.module_scope
+
+    def clear_all(self) -> None:
+        """Convenience method to clear all stacks"""
+        for stack in self:
+            stack.clear()
+
+
 class RefactoringChecker(checkers.BaseTokenChecker):
     """Looks for code which can be refactored
 
@@ -273,7 +343,7 @@ class RefactoringChecker(checkers.BaseTokenChecker):
             "raise statement.",
         ),
         "R1721": (
-            "Unnecessary use of a comprehension",
+            "Unnecessary use of a comprehension, use %s instead.",
             "unnecessary-comprehension",
             "Instead of using an identity comprehension, "
             "consider using the list, dict or set constructor. "
@@ -336,6 +406,13 @@ class RefactoringChecker(checkers.BaseTokenChecker):
             "Emitted if a resource-allocating assignment or call may be replaced by a 'with' block. "
             "By using 'with' the release of the allocated resources is ensured even in the case of an exception.",
         ),
+        "R1733": (
+            "Unnecessary dictionary index lookup, use '%s' instead",
+            "unnecessary-dict-index-lookup",
+            "Emitted when iterating over the dictionary items (key-item pairs) and accessing the "
+            "value by index lookup. "
+            "The value can be accessed directly instead.",
+        ),
     }
     options = (
         (
@@ -365,6 +442,7 @@ class RefactoringChecker(checkers.BaseTokenChecker):
     def __init__(self, linter=None):
         checkers.BaseTokenChecker.__init__(self, linter)
         self._return_nodes = {}
+        self._consider_using_with_stack = ConsiderUsingWithStack()
         self._init()
         self._never_returning_functions = None
 
@@ -374,6 +452,7 @@ class RefactoringChecker(checkers.BaseTokenChecker):
         self._nested_blocks_msg = None
         self._reported_swap_nodes = set()
         self._can_simplify_bool_op = False
+        self._consider_using_with_stack.clear_all()
 
     def open(self):
         # do this in open since config not fully initialized in __init__
@@ -492,7 +571,12 @@ class RefactoringChecker(checkers.BaseTokenChecker):
                 if self.linter.is_message_enabled("trailing-comma-tuple"):
                     self.add_message("trailing-comma-tuple", line=token.start[0])
 
+    @utils.check_messages("consider-using-with")
     def leave_module(self, _):
+        # check for context managers that have been created but not used
+        self._emit_consider_using_with_if_needed(
+            self._consider_using_with_stack.module_scope
+        )
         self._init()
 
     @utils.check_messages("too-many-nested-blocks")
@@ -523,9 +607,14 @@ class RefactoringChecker(checkers.BaseTokenChecker):
                     args=(name_node.name,),
                 )
 
-    @utils.check_messages("redefined-argument-from-local", "too-many-nested-blocks")
+    @utils.check_messages(
+        "redefined-argument-from-local",
+        "too-many-nested-blocks",
+        "unnecessary-dict-index-lookup",
+    )
     def visit_for(self, node):
         self._check_nested_blocks(node)
+        self._check_unnecessary_dict_index_lookup(node)
 
         for name in node.target.nodes_of_class(astroid.AssignName):
             self._check_redefined_argument_from_local(name)
@@ -537,7 +626,14 @@ class RefactoringChecker(checkers.BaseTokenChecker):
 
     @utils.check_messages("redefined-argument-from-local")
     def visit_with(self, node):
-        for _, names in node.items:
+        for var, names in node.items:
+            if isinstance(var, astroid.Name):
+                for stack in self._consider_using_with_stack:
+                    # We don't need to restrict the stacks we search to the current scope and outer scopes,
+                    # as e.g. the function_scope stack will be empty when we check a ``with`` on the class level.
+                    if var.name in stack:
+                        del stack[var.name]
+                        break
             if not names:
                 continue
             for name in names.nodes_of_class(astroid.AssignName):
@@ -581,10 +677,10 @@ class RefactoringChecker(checkers.BaseTokenChecker):
 
     @staticmethod
     def _type_and_name_are_equal(node_a, node_b):
-        for _type in [astroid.Name, astroid.AssignName]:
-            if all(isinstance(_node, _type) for _node in [node_a, node_b]):
+        for _type in (astroid.Name, astroid.AssignName):
+            if all(isinstance(_node, _type) for _node in (node_a, node_b)):
                 return node_a.name == node_b.name
-        if all(isinstance(_node, astroid.Const) for _node in [node_a, node_b]):
+        if all(isinstance(_node, astroid.Const) for _node in (node_a, node_b)):
             return node_a.value == node_b.value
         return False
 
@@ -609,10 +705,7 @@ class RefactoringChecker(checkers.BaseTokenChecker):
             return False
 
         # The subscript's slice needs to be the same as the test variable.
-        # Python 3.9 we no longer have the `Index` node.
         slice_value = stmt.value.slice
-        if isinstance(slice_value, astroid.Index):
-            slice_value = slice_value.value
         if not (
             self._type_and_name_are_equal(stmt.value.value, node.test.ops[0][1])
             and self._type_and_name_are_equal(slice_value, node.test.left)
@@ -765,9 +858,12 @@ class RefactoringChecker(checkers.BaseTokenChecker):
         self.add_message("simplifiable-if-expression", node=node, args=(reduced_to,))
 
     @utils.check_messages(
-        "too-many-nested-blocks", "inconsistent-return-statements", "useless-return"
+        "too-many-nested-blocks",
+        "inconsistent-return-statements",
+        "useless-return",
+        "consider-using-with",
     )
-    def leave_functiondef(self, node):
+    def leave_functiondef(self, node: astroid.FunctionDef) -> None:
         # check left-over nested blocks stack
         self._emit_nested_blocks_message_if_needed(self._nested_blocks)
         # new scope = reinitialize the stack of nested blocks
@@ -777,6 +873,19 @@ class RefactoringChecker(checkers.BaseTokenChecker):
         # check for single return or return None at the end
         self._check_return_at_the_end(node)
         self._return_nodes[node.name] = []
+        # check for context managers that have been created but not used
+        self._emit_consider_using_with_if_needed(
+            self._consider_using_with_stack.function_scope
+        )
+        self._consider_using_with_stack.function_scope.clear()
+
+    @utils.check_messages("consider-using-with")
+    def leave_classdef(self, _: astroid.ClassDef) -> None:
+        # check for context managers that have been created but not used
+        self._emit_consider_using_with_if_needed(
+            self._consider_using_with_stack.class_scope
+        )
+        self._consider_using_with_stack.class_scope.clear()
 
     @utils.check_messages("stop-iteration-return")
     def visit_raise(self, node):
@@ -862,7 +971,7 @@ class RefactoringChecker(checkers.BaseTokenChecker):
         self._check_quit_exit_call(node)
         self._check_super_with_arguments(node)
         self._check_consider_using_generator(node)
-        self._check_consider_using_with_instead_call(node)
+        self._check_consider_using_with(node)
 
     @staticmethod
     def _has_exit_in_scope(scope):
@@ -968,6 +1077,10 @@ class RefactoringChecker(checkers.BaseTokenChecker):
                 args=(len(nested_blocks), self.config.max_nested_blocks),
             )
 
+    def _emit_consider_using_with_if_needed(self, stack: Dict[str, astroid.NodeNG]):
+        for node in stack.values():
+            self.add_message("consider-using-with", node=node)
+
     @staticmethod
     def _duplicated_isinstance_types(node):
         """Get the duplicated types from the underlying isinstance calls.
@@ -1235,9 +1348,17 @@ class RefactoringChecker(checkers.BaseTokenChecker):
         "consider-swap-variables",
         "consider-using-with",
     )
-    def visit_assign(self, node):
+    def visit_assign(self, node: astroid.Assign) -> None:
+        self._append_context_managers_to_stack(node)
+        self.visit_return(node)  # remaining checks are identical as for return nodes
+
+    @utils.check_messages(
+        "simplify-boolean-expression",
+        "consider-using-ternary",
+        "consider-swap-variables",
+    )
+    def visit_return(self, node: astroid.Return) -> None:
         self._check_swap_variables(node)
-        self._check_consider_using_with_instead_assign(node)
         if self._is_and_or_ternary(node.value):
             cond, truth_value, false_value = self._and_or_ternary_arguments(node.value)
         else:
@@ -1266,18 +1387,71 @@ class RefactoringChecker(checkers.BaseTokenChecker):
             )
         self.add_message(message, node=node, args=(suggestion,))
 
-    visit_return = visit_assign
+    def _append_context_managers_to_stack(self, node: astroid.Assign) -> None:
+        if _is_inside_context_manager(node):
+            # if we are inside a context manager itself, we assume that it will handle the resource management itself.
+            return
+        if isinstance(node.targets[0], (astroid.Tuple, astroid.List, astroid.Set)):
+            assignees = node.targets[0].elts
+            value = utils.safe_infer(node.value)
+            if value is None or not hasattr(value, "elts"):
+                # We cannot deduce what values are assigned, so we have to skip this
+                return
+            values = value.elts
+        else:
+            assignees = [node.targets[0]]
+            values = [node.value]
+        if Uninferable in (assignees, values):
+            return
+        for assignee, value in zip(assignees, values):
+            if not isinstance(value, astroid.Call):
+                continue
+            inferred = utils.safe_infer(value.func)
+            if (
+                not inferred
+                or inferred.qname() not in CALLS_RETURNING_CONTEXT_MANAGERS
+                or not isinstance(assignee, (astroid.AssignName, astroid.AssignAttr))
+            ):
+                continue
+            stack = self._consider_using_with_stack.get_stack_for_frame(node.frame())
+            varname = (
+                assignee.name
+                if isinstance(assignee, astroid.AssignName)
+                else assignee.attrname
+            )
+            if varname in stack:
+                # variable was redefined before it was used in a ``with`` block
+                self.add_message(
+                    "consider-using-with",
+                    node=stack[varname],
+                )
+            stack[varname] = value
 
-    def _check_consider_using_with_instead_assign(self, node: astroid.Assign):
-        assigned = node.value
-        if isinstance(assigned, astroid.Call):
-            inferred = utils.safe_infer(assigned.func)
-            if inferred and inferred.qname() in CALLS_RETURNING_CONTEXT_MANAGERS:
-                self.add_message("consider-using-with", node=node)
-
-    def _check_consider_using_with_instead_call(self, node: astroid.Call):
+    def _check_consider_using_with(self, node: astroid.Call):
+        if _is_inside_context_manager(node):
+            # if we are inside a context manager itself, we assume that it will handle the resource management itself.
+            return
+        if (
+            node
+            in self._consider_using_with_stack.get_stack_for_frame(
+                node.frame()
+            ).values()
+        ):
+            # the result of this call was already assigned to a variable and will be checked when leaving the scope.
+            return
         inferred = utils.safe_infer(node.func)
-        if inferred and inferred.qname() in CALLS_THAT_COULD_BE_REPLACED_BY_WITH:
+        if not inferred:
+            return
+        could_be_used_in_with = (
+            # things like ``lock.acquire()``
+            inferred.qname() in CALLS_THAT_COULD_BE_REPLACED_BY_WITH
+            or (
+                # things like ``open("foo")`` which are not already inside a ``with`` statement
+                inferred.qname() in CALLS_RETURNING_CONTEXT_MANAGERS
+                and not _is_part_of_with_items(node)
+            )
+        )
+        if could_be_used_in_with and not _will_be_released_automatically(node):
             self.add_message("consider-using-with", node=node)
 
     def _check_consider_using_join(self, aug_assign):
@@ -1317,11 +1491,12 @@ class RefactoringChecker(checkers.BaseTokenChecker):
     def visit_augassign(self, node):
         self._check_consider_using_join(node)
 
-    @utils.check_messages("unnecessary-comprehension")
-    def visit_comprehension(self, node):
+    @utils.check_messages("unnecessary-comprehension", "unnecessary-dict-index-lookup")
+    def visit_comprehension(self, node: astroid.Comprehension) -> None:
         self._check_unnecessary_comprehension(node)
+        self._check_unnecessary_dict_index_lookup(node)
 
-    def _check_unnecessary_comprehension(self, node):
+    def _check_unnecessary_comprehension(self, node: astroid.Comprehension) -> None:
         if (
             isinstance(node.parent, astroid.GeneratorExp)
             or len(node.ifs) != 0
@@ -1367,7 +1542,38 @@ class RefactoringChecker(checkers.BaseTokenChecker):
         else:
             return
         if expr_list == target_list != []:
-            self.add_message("unnecessary-comprehension", node=node)
+            args: Optional[Tuple[str]] = None
+            inferred = utils.safe_infer(node.iter)
+            if isinstance(node.parent, astroid.DictComp) and isinstance(
+                inferred, astroid.objects.DictItems
+            ):
+                args = (f"{node.iter.func.expr.as_string()}",)
+            elif (
+                isinstance(node.parent, astroid.ListComp)
+                and isinstance(inferred, astroid.List)
+            ) or (
+                isinstance(node.parent, astroid.SetComp)
+                and isinstance(inferred, astroid.Set)
+            ):
+                args = (f"{node.iter.as_string()}",)
+            if args:
+                self.add_message("unnecessary-comprehension", node=node, args=args)
+                return
+
+            if isinstance(node.parent, astroid.DictComp):
+                func = "dict"
+            elif isinstance(node.parent, astroid.ListComp):
+                func = "list"
+            elif isinstance(node.parent, astroid.SetComp):
+                func = "set"
+            else:
+                return
+
+            self.add_message(
+                "unnecessary-comprehension",
+                node=node,
+                args=(f"{func}({node.iter.as_string()})",),
+            )
 
     @staticmethod
     def _is_and_or_ternary(node):
@@ -1588,3 +1794,110 @@ class RefactoringChecker(checkers.BaseTokenChecker):
             # return None"
             elif isinstance(last.value, astroid.Const) and (last.value.value is None):
                 self.add_message("useless-return", node=node)
+
+    def _check_unnecessary_dict_index_lookup(
+        self, node: Union[astroid.For, astroid.Comprehension]
+    ) -> None:
+        """Add message when accessing dict values by index lookup."""
+        # Verify that we have a .items() call and
+        # that the object which is iterated is used as a subscript in the
+        # body of the for.
+        # Is it a proper items call?
+        if (
+            isinstance(node.iter, astroid.Call)
+            and isinstance(node.iter.func, astroid.Attribute)
+            and node.iter.func.attrname == "items"
+        ):
+            inferred = utils.safe_infer(node.iter.func)
+            if not isinstance(inferred, astroid.BoundMethod):
+                return
+            iterating_object_name = node.iter.func.expr.as_string()
+
+            # Verify that the body of the for loop uses a subscript
+            # with the object that was iterated. This uses some heuristics
+            # in order to make sure that the same object is used in the
+            # for body.
+
+            children = (
+                node.body
+                if isinstance(node, astroid.For)
+                else node.parent.get_children()
+            )
+            for child in children:
+                for subscript in child.nodes_of_class(astroid.Subscript):
+                    subscript = cast(astroid.Subscript, subscript)
+
+                    if not isinstance(
+                        subscript.value, (astroid.Name, astroid.Attribute)
+                    ):
+                        continue
+
+                    value = subscript.slice
+
+                    if isinstance(node, astroid.For) and (
+                        isinstance(subscript.parent, astroid.Assign)
+                        and subscript in subscript.parent.targets
+                        or isinstance(subscript.parent, astroid.AugAssign)
+                        and subscript == subscript.parent.target
+                    ):
+                        # Ignore this subscript if it is the target of an assignment
+                        # Early termination; after reassignment dict index lookup will be necessary
+                        return
+
+                    # Case where .items is assigned to k,v (i.e., for k, v in d.items())
+                    if isinstance(value, astroid.Name):
+                        if (
+                            not isinstance(node.target, astroid.Tuple)
+                            or value.name != node.target.elts[0].name
+                            or iterating_object_name != subscript.value.as_string()
+                        ):
+                            continue
+
+                        if (
+                            isinstance(node, astroid.For)
+                            and value.lookup(value.name)[1][-1].lineno > node.lineno
+                        ):
+                            # Ignore this subscript if it has been redefined after
+                            # the for loop. This checks for the line number using .lookup()
+                            # to get the line number where the iterating object was last
+                            # defined and compare that to the for loop's line number
+                            continue
+
+                        self.add_message(
+                            "unnecessary-dict-index-lookup",
+                            node=subscript,
+                            args=(node.target.elts[1].as_string()),
+                        )
+
+                    # Case where .items is assigned to single var (i.e., for item in d.items())
+                    elif isinstance(value, astroid.Subscript):
+                        if (
+                            not isinstance(node.target, astroid.AssignName)
+                            or node.target.name != value.value.name
+                            or iterating_object_name != subscript.value.as_string()
+                        ):
+                            continue
+
+                        if (
+                            isinstance(node, astroid.For)
+                            and value.value.lookup(value.value.name)[1][-1].lineno
+                            > node.lineno
+                        ):
+                            # Ignore this subscript if it has been redefined after
+                            # the for loop. This checks for the line number using .lookup()
+                            # to get the line number where the iterating object was last
+                            # defined and compare that to the for loop's line number
+                            continue
+
+                        # check if subscripted by 0 (key)
+                        inferred = utils.safe_infer(value.slice)
+                        if (
+                            not isinstance(inferred, astroid.Const)
+                            or inferred.value != 0
+                        ):
+                            continue
+                        self.add_message(
+                            "unnecessary-dict-index-lookup",
+                            node=subscript,
+                            args=("1".join(value.as_string().rsplit("0", maxsplit=1)),),
+                        )
