@@ -42,6 +42,7 @@
 # Copyright (c) 2020 Slavfox <slavfoxman@gmail.com>
 # Copyright (c) 2020 Anthony Sottile <asottile@umich.edu>
 # Copyright (c) 2021 Marc Mueller <30130371+cdce8p@users.noreply.github.com>
+# Copyright (c) 2021 Matus Valo <matusvalo@users.noreply.github.com>
 # Copyright (c) 2021 Lorena B <46202743+lorena-b@users.noreply.github.com>
 # Copyright (c) 2021 yushao2 <36848472+yushao2@users.noreply.github.com>
 
@@ -71,8 +72,10 @@ from typing import (
 
 import _string
 import astroid
+import astroid.objects
+
+from pylint.constants import BUILTINS
 
-BUILTINS_NAME = builtins.__name__
 COMP_NODE_TYPES = (
     astroid.ListComp,
     astroid.SetComp,
@@ -301,7 +304,7 @@ def get_all_elements(
 
 def is_super(node: astroid.node_classes.NodeNG) -> bool:
     """return True if the node is referencing the "super" builtin function"""
-    if getattr(node, "name", None) == "super" and node.root().name == BUILTINS_NAME:
+    if getattr(node, "name", None) == "super" and node.root().name == BUILTINS:
         return True
     return False
 
@@ -317,7 +320,7 @@ SPECIAL_BUILTINS = ("__builtins__",)  # '__path__', '__file__')
 
 def is_builtin_object(node: astroid.node_classes.NodeNG) -> bool:
     """Returns True if the given node is an object from the __builtin__ module."""
-    return node and node.root().name == BUILTINS_NAME
+    return node and node.root().name == BUILTINS
 
 
 def is_builtin(name: str) -> bool:
@@ -791,11 +794,24 @@ def _is_property_decorator(decorator: astroid.Name) -> bool:
             if inferred.qname() in ("builtins.property", "functools.cached_property"):
                 return True
             for ancestor in inferred.ancestors():
-                if (
-                    ancestor.name == "property"
-                    and ancestor.root().name == BUILTINS_NAME
-                ):
+                if ancestor.name == "property" and ancestor.root().name == BUILTINS:
                     return True
+        elif isinstance(inferred, astroid.FunctionDef):
+            # If decorator is function, check if it has exactly one return
+            # and the return is itself a function decorated with property
+            returns: List[astroid.Return] = list(
+                inferred._get_return_nodes_skip_functions()
+            )
+            if len(returns) == 1 and isinstance(
+                returns[0].value, (astroid.Name, astroid.Attribute)
+            ):
+                inferred = safe_infer(returns[0].value)
+                if (
+                    inferred
+                    and isinstance(inferred, astroid.objects.Property)
+                    and isinstance(inferred.function, astroid.FunctionDef)
+                ):
+                    return decorated_with_property(inferred.function)
     return False
 
 
@@ -1512,3 +1528,37 @@ def get_subscript_const_value(node: astroid.Subscript) -> astroid.Const:
         )
 
     return inferred
+
+
+def get_import_name(
+    importnode: Union[astroid.Import, astroid.ImportFrom], modname: str
+) -> str:
+    """Get a prepared module name from the given import node
+
+    In the case of relative imports, this will return the
+    absolute qualified module name, which might be useful
+    for debugging. Otherwise, the initial module name
+    is returned unchanged.
+
+    :param importnode: node representing import statement.
+    :param modname: module name from import statement.
+    :returns: absolute qualified module name of the module
+        used in import.
+    """
+    if isinstance(importnode, astroid.ImportFrom):
+        if importnode.level:
+            root = importnode.root()
+            if isinstance(root, astroid.Module):
+                modname = root.relative_to_absolute_name(
+                    modname, level=importnode.level
+                )
+    return modname
+
+
+def is_node_in_guarded_import_block(node: astroid.NodeNG) -> bool:
+    """Return True if node is part for guarded if block.
+    I.e. `sys.version_info` or `typing.TYPE_CHECKING`
+    """
+    return isinstance(node.parent, astroid.If) and (
+        node.parent.is_sys_guard() or node.parent.is_typing_guard()
+    )
