@@ -40,11 +40,13 @@
 # Copyright (c) 2020 Anthony Sottile <asottile@umich.edu>
 # Copyright (c) 2020 Ashley Whetter <ashleyw@activestate.com>
 # Copyright (c) 2021 Marc Mueller <30130371+cdce8p@users.noreply.github.com>
+# Copyright (c) 2021 Sergei Lebedev <185856+superbobry@users.noreply.github.com>
+# Copyright (c) 2021 Lorena B <46202743+lorena-b@users.noreply.github.com>
 # Copyright (c) 2021 haasea <44787650+haasea@users.noreply.github.com>
 # Copyright (c) 2021 Alexander Kapshuna <kapsh@kap.sh>
 
 # Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
-# For details: https://github.com/PyCQA/pylint/blob/master/LICENSE
+# For details: https://github.com/PyCQA/pylint/blob/main/LICENSE
 
 """variables checkers for Python code
 """
@@ -391,6 +393,11 @@ MSGS = {
         "invalid-all-object",
         "Used when an invalid (non-string) object occurs in __all__.",
     ),
+    "E0605": (
+        "Invalid format for __all__, must be tuple or list",
+        "invalid-all-format",
+        "Used when __all__ has an invalid format.",
+    ),
     "E0611": (
         "No name %r in module %r",
         "no-name-in-module",
@@ -452,11 +459,6 @@ MSGS = {
         "redefined-builtin",
         "Used when a variable or function override a built-in.",
     ),
-    "W0623": (
-        "Redefining name %r from %s in exception handler",
-        "redefine-in-handler",
-        "Used when an exception handler assigns the exception to an existing name",
-    ),
     "W0631": (
         "Using possibly undefined loop variable %r",
         "undefined-loop-variable",
@@ -726,7 +728,7 @@ class VariablesChecker(BaseChecker):
         self._postponed_evaluation_enabled = is_postponed_evaluation_enabled(node)
 
         for name, stmts in node.locals.items():
-            if utils.is_builtin(name) and not utils.is_inside_except(stmts[0]):
+            if utils.is_builtin(name):
                 if self._should_ignore_redefined_builtin(stmts[0]) or name == "__doc__":
                     continue
                 self.add_message("redefined-builtin", args=name, node=stmts[0])
@@ -737,6 +739,7 @@ class VariablesChecker(BaseChecker):
         "redefined-builtin",
         "undefined-all-variable",
         "invalid-all-object",
+        "invalid-all-format",
         "unused-variable",
     )
     def leave_module(self, node):
@@ -813,8 +816,6 @@ class VariablesChecker(BaseChecker):
             return
         globs = node.root().globals
         for name, stmt in node.items():
-            if utils.is_inside_except(stmt):
-                continue
             if name in globs and not isinstance(stmt, astroid.Global):
                 definition = globs[name][0]
                 if (
@@ -1172,12 +1173,16 @@ class VariablesChecker(BaseChecker):
                     self.add_message("undefined-variable", args=name, node=node)
 
     @utils.check_messages("no-name-in-module")
-    def visit_import(self, node):
+    def visit_import(self, node: astroid.Import) -> None:
         """check modules attribute accesses"""
         if not self._analyse_fallback_blocks and utils.is_from_fallback_block(node):
             # No need to verify this, since ImportError is already
             # handled by the client code.
             return
+        if utils.is_node_in_guarded_import_block(node) is True:
+            # Don't verify import if part of guarded import block
+            # I.e. `sys.version_info` or `typing.TYPE_CHECKING`
+            return
 
         for name, _ in node.names:
             parts = name.split(".")
@@ -1190,12 +1195,16 @@ class VariablesChecker(BaseChecker):
             self._check_module_attrs(node, module, parts[1:])
 
     @utils.check_messages("no-name-in-module")
-    def visit_importfrom(self, node):
+    def visit_importfrom(self, node: astroid.ImportFrom) -> None:
         """check modules attribute accesses"""
         if not self._analyse_fallback_blocks and utils.is_from_fallback_block(node):
             # No need to verify this, since ImportError is already
             # handled by the client code.
             return
+        if utils.is_node_in_guarded_import_block(node) is True:
+            # Don't verify import if part of guarded import block
+            # I.e. `sys.version_info` or `typing.TYPE_CHECKING`
+            return
 
         name_parts = node.modname.split(".")
         try:
@@ -1827,6 +1836,10 @@ class VariablesChecker(BaseChecker):
             self._type_annotation_names.append(type_annotation.name)
             return
 
+        if isinstance(type_annotation, astroid.Attribute):
+            self._store_type_annotation_node(type_annotation.expr)
+            return
+
         if not isinstance(type_annotation, astroid.Subscript):
             return
 
@@ -1962,6 +1975,10 @@ class VariablesChecker(BaseChecker):
         if assigned is astroid.Uninferable:
             return
 
+        if not isinstance(assigned, (astroid.Tuple, astroid.List)):
+            self.add_message("invalid-all-format", node=assigned)
+            return
+
         for elt in getattr(assigned, "elts", ()):
             try:
                 elt_name = next(elt.infer())
@@ -2100,7 +2117,6 @@ class VariablesChecker(BaseChecker):
 
         consumed = []  # [(scope_locals, consumed_key)]
         metaclass = klass.metaclass()
-
         name = None
         if isinstance(klass._metaclass, astroid.Name):
             name = klass._metaclass.name
@@ -2121,25 +2137,17 @@ class VariablesChecker(BaseChecker):
                 if found:
                     consumed.append((scope_locals, name))
                     break
-
-        if found is None and not metaclass:
-            name = None
-            if isinstance(klass._metaclass, astroid.Name):
-                name = klass._metaclass.name
-            elif (
-                isinstance(klass._metaclass, astroid.Attribute)
-                and klass._metaclass.expr
-            ):
-                name = klass._metaclass.expr.name
-
-            if name is not None:
-                if not (
-                    name in astroid.Module.scope_attrs
-                    or utils.is_builtin(name)
-                    or name in self.config.additional_builtins
-                    or name in parent_node.locals
-                ):
-                    self.add_message("undefined-variable", node=klass, args=(name,))
+        if (
+            found is None
+            and not metaclass
+            and not (
+                name in astroid.Module.scope_attrs
+                or utils.is_builtin(name)
+                or name in self.config.additional_builtins
+                or name in parent_node.locals
+            )
+        ):
+            self.add_message("undefined-variable", node=klass, args=(name,))
 
         return consumed
 
